"""The artless and small template library for server-side rendering."""

__author__ = "Peter Bro"
__version__ = "0.5.0"
__license__ = "MIT"
__all__ = ["Component", "Tag", "Template", "aread_template", "read_template"]

from asyncio import to_thread
from functools import lru_cache
from pathlib import Path
from re import compile, escape
from typing import ClassVar, Mapping, Optional, Protocol, runtime_checkable

# Void tags (https://developer.mozilla.org/en-US/docs/Glossary/Void_element),
# without closing scope.
_VOID_TAGS: Mapping[str, bool] = {
    "area": True,
    "base": True,
    "br": True,
    "col": True,
    "embed": True,
    "hr": True,
    "img": True,
    "input": True,
    "link": True,
    "meta": True,
    "source": True,
    "track": True,
    "wbr": True,
}


@runtime_checkable
class Component(Protocol):
    def view(self) -> "Tag": ...


def _read_file_raw(filename: str | Path, /) -> str:
    return open(filename, "r").read()


class Tag:
    __slots__ = ("attrs", "children", "name", "parent", "text")

    def __init__(self, name: str, /, *args) -> None:
        self.name = name.lower()
        self.attrs, self.children, self.text = self._unpack_args(*args)

        self.parent: Optional["Tag"] = None

        for child in self.children:
            if not isinstance(child, self.__class__):
                raise ValueError(f"Child {child} is not a Tag instance!")
            child.parent = self

    def __str__(self) -> str:
        tag = f"<{self.name}"

        for name, value in self.attrs.items():
            tag += f' {name}="{value}"'

        if self.name in _VOID_TAGS:
            tag += " />"
            return tag

        tag += ">"

        for child in self.children:
            tag += str(child)

        tag += f"{self.text}</{self.name}>"

        return tag

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}: {self.name!r}>"

    @property
    def is_parent(self) -> bool:
        return bool(self.children)

    @property
    def is_leaf(self) -> bool:
        return not bool(self.children)

    def add_child(self, tag: "Tag", /) -> "Tag":
        tag.parent = self
        self.children.append(tag)
        return self

    @staticmethod
    def _unpack_args(*args) -> tuple[Mapping[str, str], list["Tag"], str]:
        if not args:
            return ({}, [], "")

        if len(args) > 3:
            raise ValueError("Too many arguments in construuctor!")

        _kwargs = {}

        for _ in range(len(args)):
            value = args[_]

            if not value:
                continue

            if isinstance(value, Mapping):
                name = "attrs"
            elif isinstance(value, (list, tuple)):
                name = "children"
            elif isinstance(value, str):
                name = "text"
            else:
                raise ValueError(f"Invalid type: {type(value)}!")

            _kwargs[name] = value

        return (_kwargs.get("attrs", {}), _kwargs.get("children", []), _kwargs.get("text", ""))


class Template:
    __slots__ = ("content", "name")

    DELIMITER: ClassVar[str] = "@"

    def __init__(self, /, *, name: str | Path, content: str) -> None:
        self.name = name
        self.content = content

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}: {self.name!r}>"

    def render(self, **context) -> str:
        if not context:
            return self.content

        context = {
            f"{self.DELIMITER}{key}": str(value.view() if isinstance(value, Component) else value)
            for key, value in context.items()
        }
        context_sorted = sorted(context, key=len, reverse=True)
        context_escaped = map(escape, context_sorted)
        replacement_rx = compile("|".join(context_escaped))

        return replacement_rx.sub(lambda match: context[match.group(0)], self.content)


@lru_cache()
def read_template(filename: str | Path, /) -> Template:
    return Template(name=filename, content=_read_file_raw(filename))


async def aread_template(filename: str | Path, /) -> Template:
    return Template(name=filename, content=await to_thread(_read_file_raw, filename))



from unittest import IsolatedAsyncioTestCase, TestCase
from unittest.mock import mock_open, patch

from artless_template import (
    _VOID_TAGS,
    Component,
    Tag,
    Template,
    aread_template,
    read_template,
)


class TestConstants(TestCase):
    def test_package_constants(self):
        expected_tags = [
            "area",
            "base",
            "br",
            "col",
            "embed",
            "hr",
            "img",
            "input",
            "link",
            "meta",
            "source",
            "track",
            "wbr",
        ]
        self.assertEqual(list(sorted(_VOID_TAGS.keys())), expected_tags)


class TestTemplate(TestCase):
    def setUp(self):
        self.content = (
            "<html><head><title>@title</title></head><body><h1>@header</h1></body></html>"
        )
        self.template = Template(name="main", content=self.content)

    def test_class_attributes(self):
        expected_slots = ("content", "name")
        self.assertEqual(Template.__slots__, expected_slots)

    def test_repr(self):
        self.assertEqual(repr(self.template), "<Template: 'main'>")

    def test_render_with_regular_context(self):
        context = {"title": "Test title", "header": "Test"}
        expected_result = (
            "<html><head><title>Test title</title></head><body><h1>Test</h1></body></html>"
        )

        result = self.template.render(**context)

        self.assertEqual(result, expected_result)

    def test_render_without_empty_context(self):
        result = self.template.render()
        self.assertEqual(result, self.content)

    def test_render_context_with_component(self):
        class SomeComponent(Component):
            def view(self):
                return Tag("div", "Div text")

        template = Template(
            name="test", content="<html><head><title></title></head><body>@body</body></html>"
        )
        result = template.render(body=SomeComponent())

        self.assertEqual(
            result, "<html><head><title></title></head><body><div>Div text</div></body></html>"
        )


class TestReadTemplate(TestCase):
    def setUp(self):
        self.file_content = (
            "<html><head><title>@title</title></head><body><h1>@header</h1></body></html>"
        )
        self.fake_open = mock_open(read_data=self.file_content)

    def test_read_template(self):
        with patch("builtins.open", self.fake_open) as mock_file:
            template = read_template("templates/index.html")

        mock_file.assert_called_once_with("templates/index.html", "r")

        self.assertIsInstance(template, Template)
        self.assertEqual(template.content, self.file_content)


class TestAReadTemplate(IsolatedAsyncioTestCase):
    def setUp(self):
        self.file_content = (
            "<html><head><title>@title</title></head><body><h1>@header</h1></body></html>"
        )
        self.fake_open = mock_open(read_data=self.file_content)

    async def test_aread_template(self):
        with patch("builtins.open", self.fake_open) as mock_file:
            template = await aread_template("templates/index.html")
        mock_file.assert_called_once_with("templates/index.html", "r")
        self.assertIsInstance(template, Template)
        self.assertEqual(template.content, self.file_content)


class TestTag(TestCase):
    def setUp(self):
        self.test_cases = [
            # Only tag name
            ("div", {}, None, None, "<div></div>"),
            # Only name and attributes
            ("div", {"id": "superblock"}, None, None, '<div id="superblock"></div>'),
            # Only name and text
            ("div", None, None, "Test", "<div>Test</div>"),
            # Name, attributes and text
            (
                "div",
                {"id": "superblock"},
                None,
                "Test",
                '<div id="superblock">Test</div>',
            ),
            # Name, more attributes and text
            (
                "div",
                {
                    "id": "some-id",
                    "class": "some-class",
                    "data-field": "some data field",
                },
                None,
                "Some text",
                '<div id="some-id" class="some-class" data-field="some data field">Some text</div>',
            ),
            # Name, attributes, text and children
            (
                "div",
                {"class": "parent"},
                [Tag("div", {"class": "child"}, [Tag("span", "Span text")], "Child text")],
                "Parent text",
                (
                    '<div class="parent"><div class="child"><span>Span text</span>'
                    "Child text</div>Parent text</div>"
                ),
            ),
            # Void tag
            ("br", {}, None, None, "<br />"),
            # Void tag with attributes
            (
                "area",
                {"data-url": "https://www.w3.org/"},
                None,
                None,
                '<area data-url="https://www.w3.org/" />',
            ),
            # Void tag with text
            ("area", None, None, "Text that will not be shown", "<area />"),
            # Void tag with children
            ("area", None, [Tag("span")], None, "<area />"),
        ]

    def test_render_tags(self):
        for name, attrs, children, text, expected in self.test_cases:
            with self.subTest(name=name, expected=expected):
                self.assertEqual(str(Tag(name, attrs, children, text)), expected)

    def test_repr(self):
        tag = Tag("div", {"id": "superblock"}, [Tag("span", "Span text")], "Div text")
        self.assertEqual(tag.__repr__(), "<Tag: 'div'>")

    def test_add_child(self):
        tag = Tag("div")

        self.assertEqual(len(tag.children), 0)
        self.assertEqual(str(tag), "<div></div>")

        tag.add_child(Tag("span"))

        self.assertEqual(len(tag.children), 1)
        self.assertEqual(str(tag), "<div><span></span></div>")

    def test_is_parent(self):
        tag = Tag("div")
        self.assertFalse(tag.is_parent)

        tag.add_child(Tag("span"))
        self.assertTrue(tag.is_parent)

    def test_is_leaf(self):
        tag = Tag("div")
        self.assertTrue(tag.is_leaf)

        tag.add_child(Tag("span"))
        self.assertFalse(tag.is_leaf)
